import sys, os
import numpy as np
import scipy.io
import pandas as pd
import scipy.signal
from .. import core as nts



'''
Wrappers should be able to distinguish between raw data or matlab processed data
'''

def loadSpikeData(path, index=None, fs = 20000):
	"""
	if the path contains a folder named /Analysis, 
	the script will look into it to load either
		- SpikeData.mat saved from matlab
		- SpikeData.h5 saved from this same script
	if not, the res and clu file will be loaded 
	and an /Analysis folder will be created to save the data
	Thus, the next loading of spike times will be faster
	Notes :
		If the frequency is not givne, it's assumed 20kH
	Args:
		path : string

	Returns:
		dict, array    
	"""    
	if not os.path.exists(path):
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()    
	new_path = os.path.join(path, 'Analysis/')
	if os.path.exists(new_path):
		new_path    = os.path.join(path, 'Analysis/')
		files        = os.listdir(new_path)
		if 'SpikeData.mat' in files:
			spikedata     = scipy.io.loadmat(new_path+'SpikeData.mat')
			shank         = spikedata['shank'] - 1
			if index is None:
				shankIndex     = np.arange(len(shank))
			else:
				shankIndex     = np.where(shank == index)[0]
			spikes         = {}    
			for i in shankIndex:    
				spikes[i]     = nts.Ts(spikedata['S'][0][0][0][i][0][0][0][1][0][0][2], time_units = 's')
			a             = spikes[0].as_units('s').index.values    
			if ((a[-1]-a[0])/60.)/60. > 20. : # VERY BAD        
				spikes         = {}    
				for i in shankIndex:
					spikes[i]     = nts.Ts(spikedata['S'][0][0][0][i][0][0][0][1][0][0][2]*0.0001, time_units = 's')
			return spikes, shank
		elif 'SpikeData.h5' in files:            
			final_path = os.path.join(new_path, 'SpikeData.h5')            
			try:
				spikes = pd.read_hdf(final_path, mode='r')
				# Returning a dictionnary | can be changed to return a dataframe
				toreturn = {}
				for i,j in spikes:
					toreturn[j] = nts.Ts(t=spikes[(i,j)].replace(0,np.nan).dropna().index.values, time_units = 's')
				shank = spikes.columns.get_level_values(0).values[:,np.newaxis]
				return toreturn, shank
			except:
				spikes = pd.HDFStore(final_path, 'r')
				shanks = spikes['/shanks']
				toreturn = {}
				for j in shanks.index:
					toreturn[j] = nts.Ts(spikes['/spikes/s'+str(j)])
				shank = shanks.values
				spikes.close()
				del spikes
				return toreturn, shank
			
		else:            
			print("Couldn't find any SpikeData file in "+new_path)
			print("If clu and res files are present in "+path+", a SpikeData.h5 is going to be created")

	# Creating /Analysis/ Folder here if not already present
	if not os.path.exists(new_path): os.makedirs(new_path)
	files = os.listdir(path)
	clu_files     = np.sort([f for f in files if '.clu.' in f and f[0] != '.'])
	res_files     = np.sort([f for f in files if '.res.' in f and f[0] != '.'])
	clu1         = np.sort([int(f.split(".")[-1]) for f in clu_files])
	clu2         = np.sort([int(f.split(".")[-1]) for f in res_files])
	if len(clu_files) != len(res_files) or not (clu1 == clu2).any():
		print("Not the same number of clu and res files in "+path+"; Exiting ...")
		sys.exit()
	count = 0
	spikes = []
	basename = clu_files[0].split(".")[0]
	for i, s in zip(range(len(clu_files)),clu1):
		clu = np.genfromtxt(os.path.join(path,basename+'.clu.'+str(s)),dtype=np.int32)[1:]
		if np.max(clu)>1:
			# print(i,s)
			res = np.genfromtxt(os.path.join(path,basename+'.res.'+str(s)))
			tmp = np.unique(clu).astype(int)
			idx_clu = tmp[tmp>1]
			idx_col = np.arange(count, count+len(idx_clu))	        
			tmp = pd.DataFrame(index = np.unique(res)/fs,
								columns = pd.MultiIndex.from_product([[s],idx_col]),
								data = 0, 
								dtype = np.uint16)
			for j, k in zip(idx_clu, idx_col):
				tmp.loc[res[clu==j]/fs,(s,k)] = np.uint16(k+1)
			spikes.append(tmp)
			count+=len(idx_clu)

			# tmp2 = pd.DataFrame(index=res[clu==j]/fs, data = k+1, ))
			# spikes = pd.concat([spikes, tmp2], axis = 1)


	# Returning a dictionnary
	toreturn =  {}
	shank = []
	for s in spikes:
		shank.append(s.columns.get_level_values(0).values)
		sh = np.unique(shank[-1])[0]
		for i,j in s:
			toreturn[j] = nts.Ts(t=s[(i,j)].replace(0,np.nan).dropna().index.values, time_units = 's')

	del spikes
	shank = np.hstack(shank)

	final_path = os.path.join(new_path, 'SpikeData.h5')
	store = pd.HDFStore(final_path)
	for s in toreturn.keys():
		store.put('spikes/s'+str(s), toreturn[s].as_series())
	store.put('shanks', pd.Series(index = list(toreturn.keys()), data = shank))
	store.close()

	# OLD WAY
	# spikes = pd.concat(spikes, axis = 1)
	# spikes = spikes.fillna(0)
	# spikes = spikes.astype(np.uint16)

	# Saving SpikeData.h5
	# final_path = os.path.join(new_path, 'SpikeData.h5')
	# spikes.columns.set_names(['shank', 'neuron'], inplace=True)    
	# spikes.to_hdf(final_path, key='spikes', mode='w')

	# Returning a dictionnary
	# toreturn = {}
	# for i,j in spikes:
	# 	toreturn[j] = nts.Ts(t=spikes[(i,j)].replace(0,np.nan).dropna().index.values, time_units = 's')

	# shank = spikes.columns.get_level_values(0).values[:,np.newaxis].flatten()

	return toreturn, shank

def loadXML(path):
	"""
	path should be the folder session containing the XML file
	Function returns :
		1. the number of channels
		2. the sampling frequency of the dat file or the eeg file depending of what is present in the folder
			eeg file first if both are present or both are absent
		3. the mappings shanks to channels as a dict
	Args:
		path : string

	Returns:
		int, int, dict
	"""
	if not os.path.exists(path):
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()
	listdir = os.listdir(path)
	xmlfiles = [f for f in listdir if f.endswith('.xml')]
	if not len(xmlfiles):
		print("Folder contains no xml files; Exiting ...")
		sys.exit()
	new_path = os.path.join(path, xmlfiles[0])
	
	from xml.dom import minidom	
	xmldoc 		= minidom.parse(new_path)
	nChannels 	= xmldoc.getElementsByTagName('acquisitionSystem')[0].getElementsByTagName('nChannels')[0].firstChild.data
	fs_dat 		= xmldoc.getElementsByTagName('acquisitionSystem')[0].getElementsByTagName('samplingRate')[0].firstChild.data
	fs_eeg 		= xmldoc.getElementsByTagName('fieldPotentials')[0].getElementsByTagName('lfpSamplingRate')[0].firstChild.data	
	if os.path.splitext(xmlfiles[0])[0] +'.dat' in listdir:
		fs = fs_dat
	elif os.path.splitext(xmlfiles[0])[0] +'.eeg' in listdir:
		fs = fs_eeg
	else:
		fs = fs_eeg
	shank_to_channel = {}
	groups 		= xmldoc.getElementsByTagName('anatomicalDescription')[0].getElementsByTagName('channelGroups')[0].getElementsByTagName('group')
	for i in range(len(groups)):
		shank_to_channel[i] = np.sort([int(child.firstChild.data) for child in groups[i].getElementsByTagName('channel')])
	return int(nChannels), int(fs), shank_to_channel

def makeEpochs(path, order, file = None, start=None, end = None, time_units = 's'):
	"""
	The pre-processing pipeline should spit out a csv file containing all the successive epoch of sleep/wake
	This function will load the csv and write neuroseries.IntervalSet of wake and sleep in /Analysis/BehavEpochs.h5
	If no csv exists, it's still possible to give by hand the start and end of the epochs
	Notes:
		The function assumes no header on the csv file
	Args:
		path: string
		order: list
		file: string
		start: list/array (optional)
		end: list/array (optional)
		time_units: string (optional)
	Return: 
		none
	"""		
	if not os.path.exists(path):
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()	
	if file:
		listdir 	= os.listdir(path)	
		if file not in listdir:
			print("The file "+file+" cannot be found in the path "+path)
			sys.exit()			
		filepath 	= os.path.join(path, file)
		epochs 		= pd.read_csv(filepath, header = None)
	elif file is None and len(start) and len(end):
		epochs = pd.DataFrame(np.vstack((start, end)).T)
	elif file is None and start is None and end is None:
		print("You have to specify either a file or arrays of start and end; Exiting ...")
		sys.exit()
	
	# Creating /Analysis/ Folder here if not already present
	new_path	= os.path.join(path, 'Analysis/')
	if not os.path.exists(new_path): os.makedirs(new_path)
	# Writing to BehavEpochs.h5
	new_file 	= os.path.join(new_path, 'BehavEpochs.h5')
	store 		= pd.HDFStore(new_file, 'a')
	epoch 		= np.unique(order)
	for i, n in enumerate(epoch):
		idx = np.where(np.array(order) == n)[0]
		ep = nts.IntervalSet(start = epochs.loc[idx,0],
							end = epochs.loc[idx,1],
							time_units = time_units)
		store[n] = pd.DataFrame(ep)
	store.close()

	return None

def makePositions(path, file_order, episodes, n_channels=1, trackchannel=0, names = ['ry', 'rx', 'rz', 'x', 'y', 'z'], update_wake_epoch = True):
	"""
	Assuming that makeEpochs has been runned and a file BehavEpochs.h5 can be 
	found in /Analysis/, this function will look into path  for analogin file 
	containing the TTL pulses. The position time for all events will thus be
	updated and saved in Analysis/Position.h5.
	BehavEpochs.h5 will although be updated to match the time between optitrack
	and intan
	
	Notes:
		The function assumes headers on the csv file of the position in the following order:
			['ry', 'rx', 'rz', 'x', 'y', 'z']
	Args:
		path: string
		file_order: list
		names: list
	Return: 
		None
	""" 
	if not os.path.exists(path):
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()
	files = os.listdir(path)
	for f in file_order:
		if not np.any([f+'.csv' in g for g in files]):
			print("Could not find "+f+'.csv; Exiting ...')
			sys.exit()    
	new_path = os.path.join(path, 'Analysis/')
	if not os.path.exists(new_path): os.makedirs(new_path)                
	file_epoch = os.path.join(path, 'Analysis', 'BehavEpochs.h5')
	if os.path.exists(file_epoch):
		wake_ep = loadEpoch(path, 'wake')
	else:
		makeEpochs(path, episodes, file = 'Epoch_TS.csv')
		wake_ep = loadEpoch(path, 'wake')
	if len(wake_ep) != len(file_order):
		print("Number of wake episodes doesn't match; Exiting...")
		sys.exit()

	frames = []
	others = []

	for i, f in enumerate(file_order):
		print(i, f)
		csv_file = os.path.join(path, "".join(s for s in files if f+'.csv' in s))
		position = pd.read_csv(csv_file, header = [4,5], index_col = 1)
		if 1 in position.columns:
			position = position.drop(labels = 1, axis = 1)
		position = position[~position.index.duplicated(keep='first')]
		analogin_file = os.path.splitext(csv_file)[0]+'_analogin.dat'
		if not os.path.split(analogin_file)[1] in files:
			print("No analogin.dat file found.")
			print("Please provide it as "+os.path.split(analogin_file)[1])
			print("Exiting ...")
			sys.exit()
		else:
			ttl = loadTTLPulse(analogin_file, n_channels, trackchannel)
		
		if len(ttl):
			length = np.minimum(len(ttl), len(position))
			ttl = ttl.iloc[0:length]
			position = position.iloc[0:length]
			time_offset = wake_ep.as_units('s').iloc[i,0] + ttl.index[0]
		else:
			print("No ttl for ", i, f)
			time_offset = wake_ep.as_units('s').iloc[i,0]
		
		position.index += time_offset
		wake_ep.iloc[i,0] = np.int64(np.maximum(wake_ep.as_units('s').iloc[i,0], position.index[0])*1e6)
		wake_ep.iloc[i,1] = np.int64(np.minimum(wake_ep.as_units('s').iloc[i,1], position.index[-1])*1e6)

		if len(position.columns) > 6:
			frames.append(position.iloc[:,0:6])
			others.append(position.iloc[:,6:])
		else:
			frames.append(position)
	
	position = pd.concat(frames)
	#position = nts.TsdFrame(t = position.index.values, d = position.values, time_units = 's', columns = names)
	position.columns = names
	position[['ry', 'rx', 'rz']] *= (np.pi/180)
	position[['ry', 'rx', 'rz']] += 2*np.pi
	position[['ry', 'rx', 'rz']] %= 2*np.pi
	
	if len(others):
		others = pd.concat(others)
		others.columns = names
		others[['ry', 'rx', 'rz']] *= (np.pi/180)
		others[['ry', 'rx', 'rz']] += 2*np.pi
		others[['ry', 'rx', 'rz']] %= 2*np.pi

	if update_wake_epoch:
		store = pd.HDFStore(file_epoch, 'a')
		store['wake'] = pd.DataFrame(wake_ep)
		store.close()
	
	position_file = os.path.join(path, 'Analysis', 'Position.h5')
	store = pd.HDFStore(position_file, 'w')
	store['position'] = position
	store.close()
	
	if len(others):
		walls_file = os.path.join(path, 'Analysis', 'Walls.h5')
		store = pd.HDFStore(walls_file, 'w')
		store['position'] = others
		store.close()

	return

def loadEpoch(path, epoch, episodes = None):
	"""
	load the epoch contained in path	
	If the path contains a folder analysis, the function will load either the BehavEpochs.mat or the BehavEpochs.h5
	Run makeEpochs(data_directory, ['sleep', 'wake', 'sleep', 'wake'], file='Epoch_TS.csv') to create the BehavEpochs.h5

	Args:
		path: string
		epoch: string

	Returns:
		neuroseries.IntervalSet
	"""			
	if not os.path.exists(path): # Check for path
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()
	if epoch in ['sws', 'rem']: 		
		# loading the .epoch.evt file
		file = os.path.join(path,os.path.basename(path)+'.'+epoch+'.evt')
		if os.path.exists(file):
			tmp = np.genfromtxt(file)[:,0]
			tmp = tmp.reshape(len(tmp)//2,2)/1000
			ep = nts.IntervalSet(start = tmp[:,0], end = tmp[:,1], time_units = 's')
			# TO make sure it's only in sleep since using the TheStateEditor
			sleep_ep = loadEpoch(path, 'sleep')
			ep = sleep_ep.intersect(ep)
			return ep
		else:
			print("The file ", file, "does not exist; Exiting ...")
			sys.exit()
	elif epoch == 'wake.evt.theta':
		file = os.path.join(path,os.path.basename(path)+'.'+epoch)
		if os.path.exists(file):
			tmp = np.genfromtxt(file)[:,0]
			tmp = tmp.reshape(len(tmp)//2,2)/1000
			ep = nts.IntervalSet(start = tmp[:,0], end = tmp[:,1], time_units = 's')
			return ep
		else:
			print("The file ", file, "does not exist; Exiting ...")
	filepath 	= os.path.join(path, 'Analysis')
	if os.path.exists(filepath): # Check for path/Analysis/	
		listdir		= os.listdir(filepath)
		file 		= [f for f in listdir if 'BehavEpochs' in f]
	if len(file) == 0: # Running makeEpochs		
		makeEpochs(path, episodes, file = 'Epoch_TS.csv')
		listdir		= os.listdir(filepath)
		file 		= [f for f in listdir if 'BehavEpochs' in f]
	if file[0] == 'BehavEpochs.h5':
		new_file = os.path.join(filepath, 'BehavEpochs.h5')
		store 		= pd.HDFStore(new_file, 'r')
		if '/'+epoch in store.keys():
			ep = store[epoch]
			store.close()
			return nts.IntervalSet(ep)
		else:
			print("The file BehavEpochs.h5 does not contain the key "+epoch+"; Exiting ...")
			sys.exit()
	elif file[0] == 'BehavEpochs.mat':
		behepochs = scipy.io.loadmat(os.path.join(filepath,file[0]))
		if epoch == 'wake':
			wake_ep = np.hstack([behepochs['wakeEp'][0][0][1],behepochs['wakeEp'][0][0][2]])
			return nts.IntervalSet(wake_ep[:,0], wake_ep[:,1], time_units = 's').drop_short_intervals(0.0)
		elif epoch == 'sleep':
			sleep_pre_ep, sleep_post_ep = [], []
			if 'sleepPreEp' in behepochs.keys():
				sleep_pre_ep = behepochs['sleepPreEp'][0][0]
				sleep_pre_ep = np.hstack([sleep_pre_ep[1],sleep_pre_ep[2]])
				sleep_pre_ep_index = behepochs['sleepPreEpIx'][0]
			if 'sleepPostEp' in behepochs.keys():
				sleep_post_ep = behepochs['sleepPostEp'][0][0]
				sleep_post_ep = np.hstack([sleep_post_ep[1],sleep_post_ep[2]])
				sleep_post_ep_index = behepochs['sleepPostEpIx'][0]
			if len(sleep_pre_ep) and len(sleep_post_ep):
				sleep_ep = np.vstack((sleep_pre_ep, sleep_post_ep))
			elif len(sleep_pre_ep):
				sleep_ep = sleep_pre_ep
			elif len(sleep_post_ep):
				sleep_ep = sleep_post_ep						
			return nts.IntervalSet(sleep_ep[:,0], sleep_ep[:,1], time_units = 's')
		###################################
		# WORKS ONLY FOR MATLAB FROM HERE #
		###################################		
		elif epoch == 'sws':
			sampling_freq = 1250
			new_listdir = os.listdir(path)
			for f in new_listdir:
				if 'sts.SWS' in f:
					sws = np.genfromtxt(os.path.join(path,f))/float(sampling_freq)
					return nts.IntervalSet.drop_short_intervals(nts.IntervalSet(sws[:,0], sws[:,1], time_units = 's'), 0.0)

				elif '-states.mat' in f:
					sws = scipy.io.loadmat(os.path.join(path,f))['states'][0]
					index = np.logical_or(sws == 2, sws == 3)*1.0
					index = index[1:] - index[0:-1]
					start = np.where(index == 1)[0]+1
					stop = np.where(index == -1)[0]
					return nts.IntervalSet.drop_short_intervals(nts.IntervalSet(start, stop, time_units = 's', expect_fix=True), 0.0)

		elif epoch == 'rem':
			sampling_freq = 1250
			new_listdir = os.listdir(path)
			for f in new_listdir:
				if 'sts.REM' in f:
					rem = np.genfromtxt(os.path.join(path,f))/float(sampling_freq)
					return nts.IntervalSet(rem[:,0], rem[:,1], time_units = 's').drop_short_intervals(0.0)

				elif '-states/m' in listdir:
					rem = scipy.io.loadmat(path+f)['states'][0]
					index = (rem == 5)*1.0
					index = index[1:] - index[0:-1]
					start = np.where(index == 1)[0]+1
					stop = np.where(index == -1)[0]
					return nts.IntervalSet(start, stop, time_units = 's', expect_fix=True).drop_short_intervals(0.0)

def loadPosition(path, events = None, episodes = None, n_channels=1,trackchannel=0):
	"""
	load the position contained in /Analysis/Position.h5

	Notes:
		The order of the columns is assumed to be
			['ry', 'rx', 'rz', 'x', 'y', 'z']
	Args:
		path: string
		
	Returns:
		neuroseries.TsdFrame
	"""        
	if not os.path.exists(path): # Checking for path
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()
	new_path = os.path.join(path, 'Analysis')
	if not os.path.exists(new_path): os.mkdir(new_path)
	file = os.path.join(path, 'Analysis', 'Position.h5')
	if not os.path.exists(file):
		makePositions(path, events, episodes, n_channels, trackchannel)
	if os.path.exists(file):
		store = pd.HDFStore(file, 'r')
		position = store['position']
		store.close()
		position = nts.TsdFrame(t = position.index.values, d = position.values, columns = position.columns, time_units = 's')
		return position
	else:
		print("Cannot find "+file+" for loading position")
		sys.exit()    	

def loadTTLPulse(file, n_channels = 1, channel = 0, fs = 20000):
	"""
		load ttl from analogin.dat
	"""
	f = open(file, 'rb')
	startoffile = f.seek(0, 0)
	endoffile = f.seek(0, 2)
	bytes_size = 2        
	n_samples = int((endoffile-startoffile)/n_channels/bytes_size)
	f.close()
	with open(file, 'rb') as f:
		data = np.fromfile(f, np.uint16).reshape((n_samples, n_channels))
	if n_channels == 1:
		data = data.flatten().astype(np.int32)
	else:
		data = data[:,channel].flatten().astype(np.int32)
	peaks,_ = scipy.signal.find_peaks(np.diff(data), height=30000)
	timestep = np.arange(0, len(data))/fs
	# analogin = pd.Series(index = timestep, data = data)
	peaks+=1
	ttl = pd.Series(index = timestep[peaks], data = data[peaks])    
	return ttl

def loadAuxiliary(path, n_probe = 1, fs = 20000):
	"""
	Extract the acceleration from the auxiliary.dat for each epochs
	Downsampled at 100 Hz
	Args:
		path: string
		epochs_ids: list        
	Return: 
		TsdArray
	""" 	
	if not os.path.exists(path):
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()
	if 'Acceleration.h5' in os.listdir(os.path.join(path, 'Analysis')):
		accel_file = os.path.join(path, 'Analysis', 'Acceleration.h5')
		store = pd.HDFStore(accel_file, 'r')
		accel = store['acceleration'] 
		store.close()
		accel = nts.TsdFrame(t = accel.index.values*1e6, d = accel.values) 
		return accel
	else:
		aux_files = np.sort([f for f in os.listdir(path) if 'auxiliary' in f])
		if len(aux_files)==0:
			print("Could not find "+f+'_auxiliary.dat; Exiting ...')
			sys.exit()

		accel = []
		sample_size = []
		for i, f in enumerate(aux_files):
			new_path 	= os.path.join(path, f)
			f 			= open(new_path, 'rb')
			startoffile = f.seek(0, 0)
			endoffile 	= f.seek(0, 2)
			bytes_size 	= 2
			n_samples 	= int((endoffile-startoffile)/(3*n_probe)/bytes_size)
			duration 	= n_samples/fs		
			f.close()
			tmp 		= np.fromfile(open(new_path, 'rb'), np.uint16).reshape(n_samples,3*n_probe)
			accel.append(tmp)
			sample_size.append(n_samples)
			del tmp

		accel = np.concatenate(accel)	
		factor = 37.4e-6
		# timestep = np.arange(0, len(accel))/fs
		# accel = pd.DataFrame(index = timestep, data= accel*37.4e-6)
		tmp  = []
		for i in range(accel.shape[1]):
			tmp.append(scipy.signal.resample_poly(accel[:,i]*factor, 1, 100))
		tmp = np.vstack(tmp).T
		timestep = np.arange(0, len(tmp))/(fs/100)
		tmp = pd.DataFrame(index = timestep, data = tmp)
		accel_file = os.path.join(path, 'Analysis', 'Acceleration.h5')
		store = pd.HDFStore(accel_file, 'w')
		store['acceleration'] = tmp
		store.close()
		accel = nts.TsdFrame(t = tmp.index.values*1e6, d = tmp.values) 
		return accel

def downsampleDatFile(path, n_channels = 32, fs = 20000):
	"""
	downsample .dat file to .eeg 1/16 (20000 -> 1250 Hz)
	
	Since .dat file can be very big, the strategy is to load one channel at the time,
	downsample it, and free the memory.

	Args:
		path: string
		n_channel: int
		fs: int
	Return: 
		none
	"""	
	if not os.path.exists(path):
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()
	listdir 	= os.listdir(path)
	datfile 	= os.path.basename(path) + '.dat'
	if datfile not in listdir:
		print("Folder contains no " + datfile + " file; Exiting ...")
		sys.exit()

	new_path = os.path.join(path, datfile)

	f 			= open(new_path, 'rb')
	startoffile = f.seek(0, 0)
	endoffile 	= f.seek(0, 2)
	bytes_size 	= 2
	n_samples 	= int((endoffile-startoffile)/n_channels/bytes_size)
	duration 	= n_samples/fs
	f.close()

	chunksize 	= 200000000
	eeg 		= np.zeros((int(n_samples/16),n_channels), dtype = np.int16)

	for n in range(n_channels):
		print("Ch ", n)
		# Loading		
		rawchannel = np.zeros(n_samples, np.int16)
		count = 0
		while count < n_samples:
			f 			= open(new_path, 'rb')
			seekstart 	= count*n_channels*bytes_size
			f.seek(seekstart)
			block 		= np.fromfile(f, np.int16, n_channels*np.minimum(chunksize, n_samples-count))
			f.close()
			block 		= block.reshape(np.minimum(chunksize, n_samples-count), n_channels)
			rawchannel[count:count+np.minimum(chunksize, n_samples-count)] = np.copy(block[:,n])
			count 		+= chunksize
		# Downsampling		
		eeg[:,n] 	= scipy.signal.resample_poly(rawchannel, 1, 16).astype(np.int16)
		del rawchannel
	
	# Saving
	eeg_path 	= os.path.join(path, os.path.splitext(datfile)[0]+'.eeg')
	with open(eeg_path, 'wb') as f:
		eeg.tofile(f)
		
	return

def loadUFOs(path):
	"""
	Name of the file should end with .evt.py.ufo
	"""
	import os
	name = path.split("/")[-1]
	files = os.listdir(path)
	filename = os.path.join(path, name+'.evt.py.ufo')
	if name+'.evt.py.ufo' in files:
		tmp = np.genfromtxt(path + '/' + name + '.evt.py.ufo')[:,0]
		ripples = tmp.reshape(len(tmp)//3,3)/1000
	else:
		print("No ufo in ", path)
		sys.exit()
	return (nts.IntervalSet(ripples[:,0], ripples[:,2], time_units = 's'), 
			nts.Ts(ripples[:,1], time_units = 's'))

def loadRipples(path):
	"""
	Name of the file should end with .evt.py.rip
	"""
	import os
	name = path.split("/")[-1]
	files = os.listdir(path)
	filename = os.path.join(path, name+'.evt.py.rip')
	if name+'.evt.py.rip' in files:
		tmp = np.genfromtxt(path + '/' + name + '.evt.py.rip')[:,0]
		ripples = tmp.reshape(len(tmp)//3,3)/1000
	else:
		print("No ripples in ", path)
		sys.exit()
	return (nts.IntervalSet(ripples[:,0], ripples[:,2], time_units = 's'), 
			nts.Ts(ripples[:,1], time_units = 's'))



def loadMeanWaveforms(path):
	"""
	load waveforms
	quick and dirty	
	"""
	import scipy.io
	if not os.path.exists(path):
		print("The path "+path+" doesn't exist; Exiting ...")
		sys.exit()    
	new_path = os.path.join(path, 'Analysis/')
	if os.path.exists(new_path):
		new_path    = os.path.join(path, 'Analysis/')
		files        = os.listdir(new_path)
		if 'SpikeWaveF.mat' in files:			
			# data = scipy.io.loadmat(path+'/Analysis/SpikeWaveF.mat')
			# meanWaveF = data['meanWaveF'][0]
			# maxIx = data['maxIx'][0]
			# generalinfo 	= scipy.io.loadmat(path+'/Analysis/GeneralInfo.mat')
			# shankStructure 	= loadShankStructure(generalinfo)
			# spikes,shank = loadSpikeData(path+'/Analysis/SpikeData.mat', shankStructure['thalamus'])	
			# index_neurons = [path.split("/")[-1]+"_"+str(n) for n in spikes.keys()]
			# for i, n in zip(list(spikes.keys()), index_neurons):	
			# 	to_return[n] = meanWaveF[i][maxIx[i]-1]			
			print("to test matlab")
			return
		elif "MeanWaveForms.h5" in files and "MaxWaveForms.h5" in files:
			meanwavef = pd.read_hdf(os.path.join(new_path, 'MeanWaveForms.h5'))
			maxch = pd.read_hdf(os.path.join(new_path, 'MaxWaveForms.h5'))
			return meanwavef, maxch

	# Creating /Analysis/ Folder here if not already present
	if not os.path.exists(new_path): os.makedirs(new_path)
	files = os.listdir(path)
	clu_files     = np.sort([f for f in files if 'clu' in f and f[0] != '.'])	
	spk_files	  = np.sort([f for f in files if 'spk' in f and f[0] != '.'])
	clu1         = np.sort([int(f.split(".")[-1]) for f in clu_files])
	clu2         = np.sort([int(f.split(".")[-1]) for f in spk_files])
	if len(clu_files) != len(spk_files) or not (clu1 == clu2).any():
		print("Not the same number of clu and res files in "+path+"; Exiting ...")
		sys.exit()	

	# XML INFO
	n_channels, fs, shank_to_channel 	= loadXML(path)
	from xml.dom import minidom	
	xmlfile = os.path.join(path, [f for f in files if f.endswith('.xml')][0])
	xmldoc 		= minidom.parse(xmlfile)
	nSamples 	= int(xmldoc.getElementsByTagName('nSamples')[0].firstChild.data) # assuming constant nSamples

	import xml.etree.ElementTree as ET
	root = ET.parse(xmlfile).getroot()


	count = 0
	meanwavef = []
	maxch = []
	for i, s in zip(range(len(clu_files)),clu1):
		clu = np.genfromtxt(os.path.join(path,clu_files[i]),dtype=np.int32)[1:]
		mwf = []
		mch = []		
		if np.max(clu)>1:
			# load waveforms
			file = os.path.join(path, spk_files[i])
			f = open(file, 'rb')
			startoffile = f.seek(0, 0)
			endoffile = f.seek(0, 2)
			bytes_size = 2
			n_samples = int((endoffile-startoffile)/bytes_size)
			f.close()			
			n_channel = len(root.findall('spikeDetection/channelGroups/group')[s-1].findall('channels')[0])

			data = np.memmap(file, np.int16, 'r', shape = (len(clu), nSamples, n_channel))

			#data = np.fromfile(open(file, 'rb'), np.int16)
			#data = data.reshape(len(clu),nSamples,n_channel)

			tmp = np.unique(clu).astype(int)
			idx_clu = tmp[tmp>1]
			idx_col = np.arange(count, count+len(idx_clu))	        
			for j,k in zip(idx_clu, idx_col):
				# take only a subsample of spike if too big				
				idx = np.sort(np.random.choice(np.where(clu==j)[0], 5000))
				meanw = data[idx,:,:].mean(0)
				ch = np.argmax(np.max(np.abs(meanw), 0))
				mwf.append(meanw.flatten())
				mch.append(ch)
			mwf = pd.DataFrame(np.array(mwf).T)
			mwf.columns = pd.Index(idx_col)
			mch = pd.Series(index = idx_col, data = mch)
			count += len(idx_clu)
			meanwavef.append(mwf)
			maxch.append(mch)

	meanwavef = pd.concat(meanwavef, 1)
	maxch = pd.concat(maxch)	
	meanwavef.to_hdf(os.path.join(new_path, 'MeanWaveForms.h5'), key='waveforms', mode='w')
	maxch.to_hdf(os.path.join(new_path, 'MaxWaveForms.h5'), key='channel', mode='w')
	return meanwavef, maxch

# def loadNeuronWaveform(path, index):
# 	"""
# 	load waveforms
# 	quick and dirty	
# 	"""
# 	import scipy.io
# 	if not os.path.exists(path):
# 		print("The path "+path+" doesn't exist; Exiting ...")
# 		sys.exit()    

# 	# Creating /Analysis/ Folder here if not already present	
# 	files = os.listdir(path)
# 	clu_files     = np.sort([f for f in files if 'clu' in f and f[0] != '.'])	
# 	spk_files	  = np.sort([f for f in files if 'spk' in f and f[0] != '.'])
# 	clu1         = np.sort([int(f.split(".")[-1]) for f in clu_files])
# 	clu2         = np.sort([int(f.split(".")[-1]) for f in spk_files])
# 	if len(clu_files) != len(spk_files) or not (clu1 == clu2).any():
# 		print("Not the same number of clu and res files in "+path+"; Exiting ...")
# 		sys.exit()	

# 	# XML INFO
# 	n_channels, fs, shank_to_channel 	= loadXML(path)
# 	from xml.dom import minidom	
# 	xmlfile = os.path.join(path, [f for f in files if f.endswith('.xml')][0])
# 	xmldoc 		= minidom.parse(xmlfile)
# 	nSamples 	= int(xmldoc.getElementsByTagName('nSamples')[0].firstChild.data) # assuming constant nSamples

# 	import xml.etree.ElementTree as ET
# 	root = ET.parse(xmlfile).getroot()


# 	count = 0
# 	meanwavef = []
# 	maxch = []
# 	for i, s in zip(range(len(clu_files)),clu1):
# 		clu = np.genfromtxt(os.path.join(path,clu_files[i]),dtype=np.int32)[1:]
# 		mwf = []
# 		mch = []
# 		if np.max(clu)>1:
# 			# load waveforms
# 			file = os.path.join(path, spk_files[i])
# 			f = open(file, 'rb')
# 			startoffile = f.seek(0, 0)
# 			endoffile = f.seek(0, 2)
# 			bytes_size = 2
# 			n_samples = int((endoffile-startoffile)/bytes_size)
# 			f.close()			
# 			n_channel = len(root.findall('spikeDetection/channelGroups/group')[s-1].findall('channels')[0])
# 			data = np.fromfile(open(file, 'rb'), np.int16)
# 			data = data.reshape(len(clu),nSamples,n_channel)
# 			tmp = np.unique(clu).astype(int)
# 			idx_clu = tmp[tmp>1]
# 			idx_col = np.arange(count, count+len(idx_clu))	        
# 			for j,k in zip(idx_clu, idx_col):
# 				meanw = data[clu==j].mean(0)
# 				ch = np.argmax(np.max(np.abs(meanw), 0))
# 				mwf.append(meanw.flatten())
# 				mch.append(ch)
# 			mwf = pd.DataFrame(np.array(mwf).T)
# 			mwf.columns = pd.Index(idx_col)
# 			mch = pd.Series(index = idx_col, data = mch)
# 			count += len(idx_clu)
# 			meanwavef.append(mwf)
# 			maxch.append(mch)

# 	meanwavef = pd.concat(meanwavef, 1)
# 	maxch = pd.concat(maxch)	
# 	meanwavef.to_hdf(os.path.join(new_path, 'MeanWaveForms.h5'), key='waveforms', mode='w')
# 	maxch.to_hdf(os.path.join(new_path, 'MaxWaveForms.h5'), key='channel', mode='w')
# 	return meanwavef, maxch


def loadOptoEp(path, epoch, n_channels = 2, channel = 0, fs = 20000):
	"""
		load ttl from analogin.dat
	"""
	files = os.listdir(os.path.join(path, 'Analysis'))
	if 'OptoEpochs.h5' in files:
		new_file = os.path.join(path, 'Analysis/OptoEpochs.h5')
		opto_ep = pd.read_hdf(new_file)
		return nts.IntervalSet(opto_ep)
	else:
		files = os.listdir(path)
		afile = os.path.join(path, [f for f in files if '_'+str(epoch)+'_' in f][0])
		f = open(afile, 'rb')
		startoffile = f.seek(0, 0)
		endoffile = f.seek(0, 2)
		bytes_size = 2        
		n_samples = int((endoffile-startoffile)/n_channels/bytes_size)
		f.close()
		with open(afile, 'rb') as f:
			data = np.fromfile(f, np.uint16).reshape((n_samples, n_channels))
		data = data[:,channel].flatten().astype(np.int32)

		start,_ = scipy.signal.find_peaks(np.diff(data), height=3000)
		end,_ = scipy.signal.find_peaks(np.diff(data)*-1, height=3000)
		start -= 1
		timestep = np.arange(0, len(data))/fs
		# aliging based on epoch_TS.csv
		epochs = pd.read_csv(os.path.join(path, 'Epoch_TS.csv'), header = None)
		timestep = timestep + epochs.loc[epoch,0]
		opto_ep = nts.IntervalSet(start = timestep[start], end = timestep[end], time_units = 's')
		#pd.DataFrame(opto_ep).to_hdf(os.path.join(path, 'Analysis/OptoEpochs.h5'), 'opto')
		return opto_ep	

##########################################################################################################
# TODO
##########################################################################################################

def loadShankStructure(generalinfo):
	"""
	load Shank Structure from dictionnary 
	Only useful for matlab now
	Note : 
		TODO for raw data. 

	Args:
		generalinfo : dict        

	Returns: dict		    
	"""
	shankStructure = {}
	for k,i in zip(generalinfo['shankStructure'][0][0][0][0],range(len(generalinfo['shankStructure'][0][0][0][0]))):
		if len(generalinfo['shankStructure'][0][0][1][0][i]):
			shankStructure[k[0]] = generalinfo['shankStructure'][0][0][1][0][i][0]-1
		else :
			shankStructure[k[0]] = []
	
	return shankStructure	

def loadShankMapping(path):		
	spikedata = scipy.io.loadmat(path)
	shank = spikedata['shank']
	return shank



def loadHDCellInfo(path, index):
	"""
	load the session_id_HDCells.mat file that contains the index of the HD neurons
	Only useful for matlab now
	Note : 
		TODO for raw data. 

	Args:
		generalinfo : string, array

	Returns:
		array
	"""	
	# units shoud be the value to convert in s 	
	import scipy.io
	hd_info = scipy.io.loadmat(path)['hdCellStats'][:,-1]
	return np.where(hd_info[index])[0]



def loadLFP(path, n_channels=90, channel=64, frequency=1250.0, precision='int16'):
	import neuroseries as nts	
	f = open(path, 'rb')
	startoffile = f.seek(0, 0)
	endoffile = f.seek(0, 2)
	bytes_size = 2		
	n_samples = int((endoffile-startoffile)/n_channels/bytes_size)
	duration = n_samples/frequency
	interval = 1/frequency
	f.close()
	fp = np.memmap(path, np.int16, 'r', shape = (n_samples, n_channels))		
	timestep = np.arange(0, n_samples)/frequency

	if type(channel) is not list:
		timestep = np.arange(0, n_samples)/frequency
		return nts.Tsd(timestep, fp[:,channel], time_units = 's')
	elif type(channel) is list:
		timestep = np.arange(0, n_samples)/frequency
		return nts.TsdFrame(timestep, fp[:,channel], time_units = 's')


def loadBunch_Of_LFP(path,  start, stop, n_channels=90, channel=64, frequency=1250.0, precision='int16'):
	import neuroseries as nts	
	bytes_size = 2		
	start_index = int(start*frequency*n_channels*bytes_size)
	stop_index = int(stop*frequency*n_channels*bytes_size)
	fp = np.memmap(path, np.int16, 'r', start_index, shape = (stop_index - start_index)//bytes_size)
	data = np.array(fp).reshape(len(fp)//n_channels, n_channels)

	if type(channel) is not list:
		timestep = np.arange(0, len(data))/frequency
		return nts.Tsd(timestep, data[:,channel], time_units = 's')
	elif type(channel) is list:
		timestep = np.arange(0, len(data))/frequency		
		return nts.TsdFrame(timestep, data[:,channel], time_units = 's')

def loadUpDown(path):
	import neuroseries as nts
	import os
	name = path.split("/")[-1]
	files = os.listdir(path)
	if name + '.evt.py.dow' in files:
		tmp = np.genfromtxt(path+'/'+name+'.evt.py.dow')[:,0]
		tmp = tmp.reshape(len(tmp)//2,2)/1000
		down_ep = nts.IntervalSet(start = tmp[:,0], end = tmp[:,1], time_units = 's')
	if name + '.evt.py.upp' in files:
		tmp = np.genfromtxt(path+'/'+name+'.evt.py.upp')[:,0]
		tmp = tmp.reshape(len(tmp)//2,2)/1000
		up_ep = nts.IntervalSet(start = tmp[:,0], end = tmp[:,1], time_units = 's')
	return (down_ep, up_ep)




import numpy as np
from numba import jit
import pandas as pd
import sys, os
import scipy
from scipy import signal
from itertools import combinations
#from pycircstat.descriptive import mean as circmean
#from pylab import *
from .. import core as nts



'''
Utilities functions
Feel free to add your own
'''


#########################################################
# CORRELATION
#########################################################
@jit(nopython=True)
def crossCorr(t1, t2, binsize, nbins):
	''' 
		Fast crossCorr 
	'''
	nt1 = len(t1)
	nt2 = len(t2)
	if np.floor(nbins/2)*2 == nbins:
		nbins = nbins+1

	m = -binsize*((nbins+1)/2)
	B = np.zeros(nbins)
	for j in range(nbins):
		B[j] = m+j*binsize

	w = ((nbins/2) * binsize)
	C = np.zeros(nbins)
	i2 = 1

	for i1 in range(nt1):
		lbound = t1[i1] - w
		while i2 < nt2 and t2[i2] < lbound:
			i2 = i2+1
		while i2 > 1 and t2[i2-1] > lbound:
			i2 = i2-1

		rbound = lbound
		l = i2
		for j in range(nbins):
			k = 0
			rbound = rbound+binsize
			while l < nt2 and t2[l] < rbound:
				l = l+1
				k = k+1

			C[j] += k

	# for j in range(nbins):
	# C[j] = C[j] / (nt1 * binsize)
	C = C/(nt1 * binsize/1000)

	return C

def crossCorr2(t1, t2, binsize, nbins):
	'''
		Slow crossCorr
	'''
	window = np.arange(-binsize*(nbins/2),binsize*(nbins/2)+2*binsize,binsize) - (binsize/2.)
	allcount = np.zeros(nbins+1)
	for e in t1:
		mwind = window + e
		# need to add a zero bin and an infinite bin in mwind
		mwind = np.array([-1.0] + list(mwind) + [np.max([t1.max(),t2.max()])+binsize])	
		index = np.digitize(t2, mwind)
		# index larger than 2 and lower than mwind.shape[0]-1
		# count each occurences 
		count = np.array([np.sum(index == i) for i in range(2,mwind.shape[0]-1)])
		allcount += np.array(count)
	allcount = allcount/(float(len(t1))*binsize / 1000)
	return allcount

def xcrossCorr_slow(t1, t2, binsize, nbins, nbiter, jitter, confInt):		
	times 			= np.arange(0, binsize*(nbins+1), binsize) - (nbins*binsize)/2
	H0 				= crossCorr(t1, t2, binsize, nbins)	
	H1 				= np.zeros((nbiter,nbins+1))
	t2j	 			= t2 + 2*jitter*(np.random.rand(nbiter, len(t2)) - 0.5)
	t2j 			= np.sort(t2j, 1)
	for i in range(nbiter):			
		H1[i] 		= crossCorr(t1, t2j[i], binsize, nbins)
	Hm 				= H1.mean(0)
	tmp 			= np.sort(H1, 0)
	HeI 			= tmp[int((1-confInt)/2*nbiter),:]
	HeS 			= tmp[int((confInt + (1-confInt)/2)*nbiter)]
	Hstd 			= np.std(tmp, 0)

	return (H0, Hm, HeI, HeS, Hstd, times)

def xcrossCorr_fast(t1, t2, binsize, nbins, nbiter, jitter, confInt):		
	times 			= np.arange(0, binsize*(nbins*2+1), binsize) - (nbins*2*binsize)/2
	# need to do a cross-corr of double size to convolve after and avoid boundary effect
	H0 				= crossCorr(t1, t2, binsize, nbins*2)	
	window_size 	= 2*jitter//binsize
	window 			= np.ones(window_size)*(1/window_size)
	Hm 				= np.convolve(H0, window, 'same')
	Hstd			= np.sqrt(np.var(Hm))	
	HeI 			= np.NaN
	HeS 			= np.NaN	
	return (H0, Hm, HeI, HeS, Hstd, times)	

def compute_AutoCorrs(spks, ep, binsize = 5, nbins = 200):
	# First let's prepare a pandas dataframe to receive the data
	times = np.arange(0, binsize*(nbins+1), binsize) - (nbins*binsize)/2	
	autocorrs = pd.DataFrame(index = times, columns = list(spks.keys()))
	firing_rates = pd.Series(index = list(spks.keys()))

	# Now we can iterate over the dictionnary of spikes
	for i in spks:
		# First we extract the time of spikes in ms during wake
		spk_time = spks[i].restrict(ep).as_units('ms').index.values
		# Calling the crossCorr function
		autocorrs[i] = crossCorr(spk_time, spk_time, binsize, nbins)
		# Computing the mean firing rate
		firing_rates[i] = len(spk_time)/ep.tot_length('s')

	# We can divide the autocorrs by the firing_rates
	autocorrs = autocorrs / firing_rates

	# And don't forget to replace the 0 ms for 0
	autocorrs.loc[0] = 0.0
	return autocorrs, firing_rates

def compute_CrossCorrs(spks, ep, binsize=10, nbins = 2000, norm = False):
	"""
		
	"""	
	neurons = list(spks.keys())
	times = np.arange(0, binsize*(nbins+1), binsize) - (nbins*binsize)/2
	cc = pd.DataFrame(index = times, columns = list(combinations(neurons, 2)))
		
	for i,j in cc.columns:		
		spk1 = spks[i].restrict(ep).as_units('ms').index.values
		spk2 = spks[j].restrict(ep).as_units('ms').index.values		
		tmp = crossCorr(spk1, spk2, binsize, nbins)		
		fr = len(spk2)/ep.tot_length('s')
		if norm:
			cc[(i,j)] = tmp/fr
		else:
			cc[(i,j)] = tmp
	return cc

def compute_PairCrossCorr(spks, ep, pair, binsize=10, nbins = 2000, norm = False):
	"""
		
	"""		
	times = np.arange(0, binsize*(nbins+1), binsize) - (nbins*binsize)/2	
	spk1 = spks[pair[0]].restrict(ep).as_units('ms').index.values
	spk2 = spks[pair[1]].restrict(ep).as_units('ms').index.values		
	tmp = crossCorr(spk1, spk2, binsize, nbins)		
	fr = len(spk2)/ep.tot_length('s')
	tmp = pd.Series(index = times, data = tmp)
	if norm:
		tmp = tmp/fr
	else:
		tmp = tmp
	return tmp

def compute_EventCrossCorr(spks, evt, ep, binsize = 5, nbins = 1000, norm=False):
	"""
	"""
	neurons = list(spks.keys())
	times = np.arange(0, binsize*(nbins+1), binsize) - (nbins*binsize)/2
	cc = pd.DataFrame(index = times, columns = neurons)
	tsd1 = evt.restrict(ep).as_units('ms').index.values
	for i in neurons:
		spk2 = spks[i].restrict(ep).as_units('ms').index.values
		tmp = crossCorr(tsd1, spk2, binsize, nbins)
		fr = len(spk2)/ep.tot_length('s')
		if norm:
			cc[i] = tmp/fr
		else:
			cc[i] = tmp
	return cc
		
def compute_ISI(spks, ep, maxisi, nbins, log_=False):
	"""
	"""
	neurons = list(spks.keys())
	if log_:
		bins = np.linspace(np.log10(1), np.log10(maxisi), nbins)
	else:
		bins = np.linspace(0, maxisi, nbins)
		
	isi = pd.DataFrame(index =  bins[0:-1] + np.diff(bins)/2, columns = neurons)
	for i in neurons:
		tmp = []
		for j in ep.index:
			tmp.append(np.diff(spks[i].restrict(ep.loc[[j]]).as_units('ms').index.values))
		tmp = np.hstack(tmp)
		if log_:
			isi[i], _ = np.histogram(np.log10(tmp), bins)
		else:
			isi[i], _ = np.histogram(tmp, bins)

	return isi

def compute_AllPairsCrossCorrs(spks, ep, binsize=10, nbins = 2000, norm = False):
	"""
		
	"""	
	neurons = list(spks.keys())
	times = np.arange(0, binsize*(nbins+1), binsize) - (nbins*binsize)/2
	cc = pd.DataFrame(index = times, columns = list(combinations(neurons, 2)))
		
	for i,j in cc.columns:		
		spk1 = spks[i].restrict(ep).as_units('ms').index.values
		spk2 = spks[j].restrict(ep).as_units('ms').index.values		
		tmp = crossCorr(spk1, spk2, binsize, nbins)		
		fr = len(spk2)/ep.tot_length('s')
		if norm:
			cc[(i,j)] = tmp/fr
		else:
			cc[(i,j)] = tmp
	return cc

def compute_AsyncCrossCorrs(spks, ep, binsize=10, nbins = 2000, norm = False, edge = 20):
	"""
		
	"""	
	neurons = list(spks.keys())
	times = np.arange(0, binsize*(nbins+1), binsize) - (nbins*binsize)/2
	cc = pd.DataFrame(index = times, columns = list(combinations(neurons, 2)))
		
	for i,j in cc.columns:		
		spk1 = spks[i].restrict(ep).as_units('ms').index.values
		spk2 = spks[j].restrict(ep).as_units('ms').index.values		

		spksync = []
		spkasync = []
		for t in spk2:			
			if np.sum(np.abs(t-spk1)<edge):
				spksync.append(t)
			else:
				spkasync.append(t)


		# tmp = crossCorr(spk1, spk2, binsize, nbins)		
		tmp = crossCorr(spk1, np.array(spkasync), binsize, nbins)
		fr = len(spkasync)/ep.tot_length('s')
		if norm:
			cc[(i,j)] = tmp/fr
		else:
			cc[(i,j)] = tmp
	return cc

def compute_RandomCrossCorrs(spks, ep,  binsize=10, nbins = 2000, norm = False, percent = 0.5):
	"""
		
	"""	
	neurons = list(spks.keys())
	times = np.arange(0, binsize*(nbins+1), binsize) - (nbins*binsize)/2
	cc = pd.DataFrame(index = times, columns = list(combinations(neurons, 2)))
		
	for i,j in cc.columns:		
		spk1 = spks[i].restrict(ep).as_units('ms').index.values
		spk2 = spks[j].restrict(ep).as_units('ms').index.values		

		spk1_random = np.sort(np.random.choice(spk1, int(len(spk1)*percent), replace=False))
		spk2_random = np.sort(np.random.choice(spk2, int(len(spk2)*percent), replace=False))

		# tmp = crossCorr(spk1, spk2, binsize, nbins)		
		tmp = crossCorr(spk1_random, spk2_random, binsize, nbins)
		fr = len(spk2_random)/ep.tot_length('s')
		if norm:
			cc[(i,j)] = tmp/fr
		else:
			cc[(i,j)] = tmp
	return cc

#########################################################
# VARIOUS
#########################################################
def computeLMNAngularTuningCurves(spikes, angle, ep, nb_bins = 180, frequency = 120.0, bin_size = 100):
	tmp 			= pd.Series(index = angle.index.values, data = np.unwrap(angle.values))	
	tmp2 			= tmp.rolling(window=50,win_type='gaussian',center=True,min_periods=1).mean(std=10.0)	
	bin_size 		= bin_size * 1000
	time_bins		= np.arange(tmp.index[0], tmp.index[-1]+bin_size, bin_size) # assuming microseconds
	index 			= np.digitize(tmp2.index.values, time_bins)
	tmp3 			= tmp2.groupby(index).mean()
	tmp3.index 		= time_bins[np.unique(index)-1]+bin_size/2
	tmp3 			= nts.Tsd(tmp3)
	tmp4			= np.diff(tmp3.values)/np.diff(tmp3.as_units('s').index.values)
	newangle 		= nts.Tsd(t = tmp3.index.values, d = tmp3.values%(2*np.pi))
	velocity 		= nts.Tsd(t=tmp3.index.values[1:], d = tmp4)
	velocity 		= velocity.restrict(ep)	
	velo_spikes 	= {}	
	for k in spikes: velo_spikes[k]	= velocity.realign(spikes[k].restrict(ep))
	# bins_velocity	= np.array([velocity.min(), -2*np.pi/3, -np.pi/6, np.pi/6, 2*np.pi/3, velocity.max()+0.001])
	bins_velocity	= np.array([velocity.min(), -np.pi/6, np.pi/6, velocity.max()+0.001])

	idx_velocity 	= {k:np.digitize(velo_spikes[k].values, bins_velocity)-1 for k in spikes}

	bins 			= np.linspace(0, 2*np.pi, nb_bins)
	idx 			= bins[0:-1]+np.diff(bins)/2
	tuning_curves 	= {i:pd.DataFrame(index = idx, columns = list(spikes.keys())) for i in range(3)}	

	# for i,j in zip(range(3),range(0,6,2)):
	for i,j in zip(range(3),range(3)):
		for k in spikes:
			spks 			= spikes[k].restrict(ep)			
			spks 			= spks[idx_velocity[k] == j]
			angle_spike 	= newangle.restrict(ep).realign(spks)
			spike_count, bin_edges = np.histogram(angle_spike, bins)
			tmp 			= newangle.loc[velocity.index[np.logical_and(velocity.values>bins_velocity[j], velocity.values<bins_velocity[j+1])]]
			occupancy, _ 	= np.histogram(tmp, bins)
			spike_count 	= spike_count/occupancy	
			tuning_curves[i][k] = spike_count*(1/(bin_size*1e-6))

	return tuning_curves, velocity, bins_velocity

def computeAngularTuningCurves(spikes, angle, ep, nb_bins = 180, frequency = 120.0):
	bins 			= np.linspace(0, 2*np.pi, nb_bins)
	idx 			= bins[0:-1]+np.diff(bins)/2
	tuning_curves 	= pd.DataFrame(index = idx, columns = list(spikes.keys()))	
	angle 			= angle.restrict(ep)
	# Smoothing the angle here
	tmp 			= pd.Series(index = angle.index.values, data = np.unwrap(angle.values))
	tmp2 			= tmp.rolling(window=50,win_type='gaussian',center=True,min_periods=1).mean(std=10.0)
	angle			= nts.Tsd(tmp2%(2*np.pi))
	for k in spikes:
		spks 			= spikes[k]
		# true_ep 		= nts.IntervalSet(start = np.maximum(angle.index[0], spks.index[0]), end = np.minimum(angle.index[-1], spks.index[-1]))		
		spks 			= spks.restrict(ep)	
		angle_spike 	= angle.restrict(ep).realign(spks)
		spike_count, bin_edges = np.histogram(angle_spike, bins)
		occupancy, _ 	= np.histogram(angle, bins)
		spike_count 	= spike_count/occupancy		
		tuning_curves[k] = spike_count*frequency	

	return tuning_curves

def computeSpatialInfo(tc, angle, ep):
	nb_bins = tc.shape[0]+1
	bins 	= np.linspace(0, 2*np.pi, nb_bins)
	angle 			= angle.restrict(ep)
	# Smoothing the angle here
	tmp 			= pd.Series(index = angle.index.values, data = np.unwrap(angle.values))
	tmp2 			= tmp.rolling(window=50,win_type='gaussian',center=True,min_periods=1).mean(std=10.0)
	angle			= nts.Tsd(tmp2%(2*np.pi))
	pf = tc.values
	occupancy, _ 	= np.histogram(angle, bins)
	occ = np.atleast_2d(occupancy/occupancy.sum()).T
	f = np.sum(pf * occ, 0)
	pf = pf / f
	SI = np.sum(occ * pf * np.log2(pf), 0)
	SI = pd.DataFrame(index = tc.columns, columns = ['SI'], data = SI)
	return SI


def findHDCells(tuning_curves, z = 50, p = 0.0001 , m = 1):
	"""
		Peak firing rate larger than 1
		and Rayleigh test p<0.001 & z > 100
	"""
	cond1 = tuning_curves.max()>m
	from pycircstat.tests import rayleigh
	stat = pd.DataFrame(index = tuning_curves.columns, columns = ['pval', 'z'])
	for k in tuning_curves:
		stat.loc[k] = rayleigh(tuning_curves[k].index.values, tuning_curves[k].values)
	cond2 = np.logical_and(stat['pval']<p,stat['z']>z)
	tokeep = stat.index.values[np.where(np.logical_and(cond1, cond2))[0]]
	return tokeep, stat

def decodeHD(tuning_curves, spikes, ep, bin_size = 200, px = None):
	"""
		See : Zhang, 1998, Interpreting Neuronal Population Activity by Reconstruction: Unified Framework With Application to Hippocampal Place Cells
		tuning_curves: pd.DataFrame with angular position as index and columns as neuron
		spikes : dictionnary of spike times
		ep : nts.IntervalSet, the epochs for decoding
		bin_size : in ms (default:200ms)
		px : Occupancy. If None, px is uniform
	"""		
	if len(ep) == 1:
		bins = np.arange(ep.as_units('ms').start.iloc[0], ep.as_units('ms').end.iloc[-1], bin_size)
	else:
		# ep2 = nts.IntervalSet(ep.copy().as_units('ms'))
		# ep2 = ep2.drop_short_intervals(bin_size*2)
		# bins = []
		# for i in ep2.index:
		# 	bins.append(np.arange())
		# bins = np.arange(ep2.start.iloc[0], ep.end.iloc[-1], bin_size)
		print("TODO")
		sys.exit()


	order = tuning_curves.columns.values
	# TODO CHECK MATCH

	# smoothing with a non-normalized gaussian
	w = scipy.signal.gaussian(51, 2)
	
	spike_counts = pd.DataFrame(index = bins[0:-1]+np.diff(bins)/2, columns = order)
	for n in spike_counts:		
		spks = spikes[n].restrict(ep).as_units('ms').index.values
		tmp = np.histogram(spks, bins)
		spike_counts[n] = np.convolve(tmp[0], w, mode = 'same')
		# spike_counts[k] = tmp[0]

	tcurves_array = tuning_curves.values
	spike_counts_array = spike_counts.values
	proba_angle = np.zeros((spike_counts.shape[0], tuning_curves.shape[0]))

	part1 = np.exp(-(bin_size/1000)*tcurves_array.sum(1))
	if px is not None:
		part2 = px
	else:
		part2 = np.ones(tuning_curves.shape[0])
	#part2 = np.histogram(position['ry'], np.linspace(0, 2*np.pi, 61), weights = np.ones_like(position['ry'])/float(len(position['ry'])))[0]
	
	for i in range(len(proba_angle)):
		part3 = np.prod(tcurves_array**spike_counts_array[i], 1)
		p = part1 * part2 * part3
		proba_angle[i] = p/p.sum() # Normalization process here

	proba_angle  = pd.DataFrame(index = spike_counts.index.values, columns = tuning_curves.index.values, data= proba_angle)	
	proba_angle = proba_angle.astype('float')
	decoded = nts.Tsd(t = proba_angle.index.values, d = proba_angle.idxmax(1).values, time_units = 'ms')
	return decoded, proba_angle, spike_counts

def computePlaceFields(spikes, position, ep, nb_bins = 200, frequency = 120.0):
	place_fields = {}
	position_tsd = position.restrict(ep)
	xpos = position_tsd.iloc[:,0]
	ypos = position_tsd.iloc[:,1]
	xbins = np.linspace(xpos.min(), xpos.max()+1e-6, nb_bins+1)
	ybins = np.linspace(ypos.min(), ypos.max()+1e-6, nb_bins+1)
	for n in spikes:
		position_spike = position_tsd.realign(spikes[n].restrict(ep))
		spike_count,_,_ = np.histogram2d(position_spike.iloc[:,1].values, position_spike.iloc[:,0].values, [ybins,xbins])
		occupancy, _, _ = np.histogram2d(ypos, xpos, [ybins,xbins])
		mean_spike_count = spike_count/(occupancy+1)
		place_field = mean_spike_count*frequency    
		place_fields[n] = pd.DataFrame(index = ybins[0:-1][::-1],columns = xbins[0:-1], data = place_field)
		
	extent = (xbins[0], xbins[-1], ybins[0], ybins[-1]) # USEFUL FOR MATPLOTLIB
	return place_fields, extent

def computeOccupancy(position_tsd, nb_bins = 100):
    xpos = position_tsd.iloc[:,0]
    ypos = position_tsd.iloc[:,1]    
    xbins = np.linspace(xpos.min(), xpos.max()+1e-6, nb_bins+1)
    ybins = np.linspace(ypos.min(), ypos.max()+1e-6, nb_bins+1)
    occupancy, _, _ = np.histogram2d(ypos, xpos, [ybins,xbins])
    return occupancy

def computeAngularVelocityTuningCurves(spikes, angle, ep, nb_bins = 61, bin_size = 10000, norm=True):
	tmp 			= pd.Series(index = angle.index.values, data = np.unwrap(angle.values))
	tmp2 			= tmp.rolling(window=100,win_type='gaussian',center=True,min_periods=1).mean(std=10.0)
	time_bins		= np.arange(tmp.index[0], tmp.index[-1]+bin_size, bin_size) # assuming microseconds
	index 			= np.digitize(tmp2.index.values, time_bins)
	tmp3 			= tmp2.groupby(index).mean()
	tmp3.index 		= time_bins[np.unique(index)-1]+bin_size/2
	tmp3 			= nts.Tsd(tmp3)
	tmp4			= np.diff(tmp3.values)/np.diff(tmp3.as_units('s').index.values)
	tmp2 			= nts.Tsd(tmp2)
	tmp4			= np.diff(tmp2.values)/np.diff(tmp2.as_units('s').index.values)	
	velocity 		= nts.Tsd(t=tmp2.index.values[1:], d = tmp4)
	velocity 		= velocity.restrict(ep)	
	bins 			= np.linspace(-2*np.pi, 2*np.pi, nb_bins)
	idx 			= bins[0:-1]+np.diff(bins)/2
	velo_curves		= pd.DataFrame(index = idx, columns = list(spikes.keys()))

	for k in spikes:
		spks 		= spikes[k]
		spks 		= spks.restrict(ep)
		speed_spike = velocity.realign(spks)
		spike_count, bin_edges = np.histogram(speed_spike, bins)
		occupancy, _ = np.histogram(velocity.restrict(ep), bins)
		spike_count = spike_count/(occupancy+1)
		velo_curves[k] = spike_count*(1/(bin_size*1e-6))
		# normalizing by firing rate 
		if norm:
			velo_curves[k] = velo_curves[k]/(len(spikes[k].restrict(ep))/ep.tot_length('s'))

	return velo_curves

def smoothAngularTuningCurves(tuning_curves, window = 20, deviation = 3.0):
	new_tuning_curves = {}	
	for i in tuning_curves.columns:
		tcurves = tuning_curves[i]
		offset = np.mean(np.diff(tcurves.index.values))
		padded 	= pd.Series(index = np.hstack((tcurves.index.values-(2*np.pi)-offset,
												tcurves.index.values,
												tcurves.index.values+(2*np.pi)+offset)),
							data = np.hstack((tcurves.values, tcurves.values, tcurves.values)))
		smoothed = padded.rolling(window=window,win_type='gaussian',center=True,min_periods=1).mean(std=deviation)		
		new_tuning_curves[i] = smoothed.loc[tcurves.index]

	new_tuning_curves = pd.DataFrame.from_dict(new_tuning_curves)

	return new_tuning_curves

def computeMeanFiringRate(spikes, epochs, name):
	mean_frate = pd.DataFrame(index = spikes.keys(), columns = name)
	for n, ep in zip(name, epochs):
		for k in spikes:
			mean_frate.loc[k,n] = len(spikes[k].restrict(ep))/ep.tot_length('s')
	return mean_frate

def computeSpeedTuningCurves(spikes, position, ep, bin_size = 0.1, nb_bins = 20, speed_max = 0.4):
	time_bins 	= np.arange(position.index[0], position.index[-1]+bin_size*1e6, bin_size*1e6)
	index 		= np.digitize(position.index.values, time_bins)
	tmp 		= position.groupby(index).mean()
	tmp.index 	= time_bins[np.unique(index)-1]+(bin_size*1e6)/2
	distance	= np.sqrt(np.power(np.diff(tmp['x']), 2) + np.power(np.diff(tmp['z']), 2))
	speed 		= nts.Tsd(t = tmp.index.values[0:-1]+ bin_size/2, d = distance/bin_size)
	speed 		= speed.restrict(ep)
	bins 		= np.linspace(0, speed_max, nb_bins)
	idx 		= bins[0:-1]+np.diff(bins)/2
	speed_curves = pd.DataFrame(index = idx,columns = np.arange(len(spikes)))
	for k in spikes:
		spks 	= spikes[k]
		spks 	= spks.restrict(ep)
		speed_spike = speed.realign(spks)
		spike_count, bin_edges = np.histogram(speed_spike, bins)
		occupancy, _ = np.histogram(speed, bins)
		spike_count = spike_count/(occupancy+1)
		speed_curves[k] = spike_count/bin_size

	return speed_curves

def computeAccelerationTuningCurves(spikes, position, ep, bin_size = 0.1, nb_bins = 40):
	time_bins 	= np.arange(position.index[0], position.index[-1]+bin_size*1e6, bin_size*1e6)
	index 		= np.digitize(position.index.values, time_bins)
	tmp 		= position.groupby(index).mean()
	tmp.index 	= time_bins[np.unique(index)-1]+(bin_size*1e6)/2
	distance	= np.sqrt(np.power(np.diff(tmp['x']), 2) + np.power(np.diff(tmp['z']), 2))
	speed 		= nts.Tsd(t = tmp.index.values[0:-1]+ bin_size/2, d = distance/bin_size)
	speed 		= speed.restrict(ep)
	speed 		= speed.as_series()
	speed2 		= speed.rolling(window=10, win_type='gaussian', center= True, min_periods=1).mean(std = 1.0)
	accel 		= nts.Tsd(t = speed2.index.values[0:-1] + np.diff(speed2.index.values)/2, d = np.diff(speed2.values))	
	bins 		= np.linspace(accel.min(), accel.max(), nb_bins)
	idx 		= bins[0:-1]+np.diff(bins)/2
	accel_curves = pd.DataFrame(index = idx,columns = np.arange(len(spikes)))
	for k in spikes:
		spks 	= spikes[k]
		spks 	= spks.restrict(ep)
		accel_spike = accel.realign(spks)
		spike_count, bin_edges = np.histogram(accel_spike, bins)
		occupancy, _ = np.histogram(accel, bins)
		spike_count = spike_count/(occupancy+1)
		accel_curves[k] = spike_count/bin_size

	return accel_curves

def refineSleepFromAccel(acceleration, sleep_ep):
	vl = acceleration[0].restrict(sleep_ep)
	vl = vl.as_series().diff().abs().dropna()	
	a, _ = scipy.signal.find_peaks(vl, 0.025)
	peaks = nts.Tsd(vl.iloc[a])
	duration = np.diff(peaks.as_units('s').index.values)
	interval = nts.IntervalSet(start = peaks.index.values[0:-1], end = peaks.index.values[1:])

	newsleep_ep = interval.iloc[duration>15.0]
	newsleep_ep = newsleep_ep.reset_index(drop=True)
	newsleep_ep = newsleep_ep.merge_close_intervals(100000, time_units ='us')

	newsleep_ep	= sleep_ep.intersect(newsleep_ep)

	return newsleep_ep

def splitWake(ep):
	if len(ep) != 1:
		print('Cant split wake in 2')
		sys.exit()
	tmp = np.zeros((2,2))
	tmp[0,0] = ep.values[0,0]
	tmp[1,1] = ep.values[0,1]
	tmp[0,1] = tmp[1,0] = ep.values[0,0] + np.diff(ep.values[0])/2
	return nts.IntervalSet(start = tmp[:,0], end = tmp[:,1])

def centerTuningCurves(tcurve):
	"""
	center tuning curves by peak
	"""
	peak 			= pd.Series(index=tcurve.columns,data = np.array([circmean(tcurve.index.values, tcurve[i].values) for i in tcurve.columns]))
	new_tcurve 		= []
	for p in tcurve.columns:	
		x = tcurve[p].index.values - tcurve[p].index[tcurve[p].index.get_loc(peak[p], method='nearest')]
		x[x<-np.pi] += 2*np.pi
		x[x>np.pi] -= 2*np.pi
		tmp = pd.Series(index = x, data = tcurve[p].values).sort_index()
		new_tcurve.append(tmp.values)
	new_tcurve = pd.DataFrame(index = np.linspace(-np.pi, np.pi, tcurve.shape[0]+1)[0:-1], data = np.array(new_tcurve).T, columns = tcurve.columns)
	return new_tcurve

def offsetTuningCurves(tcurve, diffs):
	"""
	offseting tuning curves synced by diff
	"""	
	new_tcurve 		= []
	for p in tcurve.columns:	
		x = tcurve[p].index.values - tcurve[p].index[tcurve[p].index.get_loc(diffs[p], method='nearest')]
		x[x<-np.pi] += 2*np.pi
		x[x>np.pi] -= 2*np.pi
		tmp = pd.Series(index = x, data = tcurve[p].values).sort_index()
		new_tcurve.append(tmp.values)
	new_tcurve = pd.DataFrame(index = np.linspace(-np.pi, np.pi, tcurve.shape[0]+1)[0:-1], data = np.array(new_tcurve).T, columns = tcurve.columns)
	return new_tcurve


#########################################################
# LFP FUNCTIONS
#########################################################
def butter_bandpass(lowcut, highcut, fs, order=5):
	from scipy.signal import butter
	nyq = 0.5 * fs
	low = lowcut / nyq
	high = highcut / nyq
	b, a = butter(order, [low, high], btype='band')
	return b, a

def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
	from scipy.signal import lfilter
	b, a = butter_bandpass(lowcut, highcut, fs, order=order)
	y = lfilter(b, a, data)
	return y

def downsample(tsd, up, down):
	import scipy.signal
	import neuroseries as nts
	dtsd = scipy.signal.resample_poly(tsd.values, up, down)
	dt = tsd.as_units('s').index.values[np.arange(0, tsd.shape[0], down)]
	if len(tsd.shape) == 1:		
		return nts.Tsd(dt, dtsd, time_units = 's')
	elif len(tsd.shape) == 2:
		return nts.TsdFrame(dt, dtsd, time_units = 's', columns = list(tsd.columns))

def getPeaksandTroughs(lfp, min_points):
	"""	 
		At 250Hz (1250/5), 2 troughs cannont be closer than 20 (min_points) points (if theta reaches 12Hz);		
	"""
	import neuroseries as nts
	import scipy.signal
	if isinstance(lfp, nts.time_series.Tsd):
		troughs 		= nts.Tsd(lfp.as_series().iloc[scipy.signal.argrelmin(lfp.values, order =min_points)[0]], time_units = 'us')
		peaks 			= nts.Tsd(lfp.as_series().iloc[scipy.signal.argrelmax(lfp.values, order =min_points)[0]], time_units = 'us')
		tmp 			= nts.Tsd(troughs.realign(peaks, align = 'next').as_series().drop_duplicates('first')) # eliminate double peaks
		peaks			= peaks[tmp.index]
		tmp 			= nts.Tsd(peaks.realign(troughs, align = 'prev').as_series().drop_duplicates('first')) # eliminate double troughs
		troughs 		= troughs[tmp.index]
		return (peaks, troughs)
	elif isinstance(lfp, nts.time_series.TsdFrame):
		peaks 			= nts.TsdFrame(lfp.index.values, np.zeros(lfp.shape))
		troughs			= nts.TsdFrame(lfp.index.values, np.zeros(lfp.shape))
		for i in lfp.keys():
			peaks[i], troughs[i] = getPeaksandTroughs(lfp[i], min_points)
		return (peaks, troughs)

def getPhase(lfp, fmin, fmax, nbins, fsamp, power = False):
	""" Continuous Wavelets Transform
		return phase of lfp in a Tsd array
	"""
	import neuroseries as nts
	from Wavelets import MyMorlet as Morlet
	if isinstance(lfp, nts.time_series.TsdFrame):
		allphase 		= nts.TsdFrame(lfp.index.values, np.zeros(lfp.shape))
		allpwr 			= nts.TsdFrame(lfp.index.values, np.zeros(lfp.shape))
		for i in lfp.keys():
			allphase[i], allpwr[i] = getPhase(lfp[i], fmin, fmax, nbins, fsamp, power = True)
		if power:
			return allphase, allpwr
		else:
			return allphase			

	elif isinstance(lfp, nts.time_series.Tsd):
		cw 				= Morlet(lfp.values, fmin, fmax, nbins, fsamp)
		cwt 			= cw.getdata()
		cwt 			= np.flip(cwt, axis = 0)
		wave 			= np.abs(cwt)**2.0
		phases 			= np.arctan2(np.imag(cwt), np.real(cwt)).transpose()	
		cwt 			= None
		index 			= np.argmax(wave, 0)
		# memory problem here, need to loop
		phase 			= np.zeros(len(index))	
		for i in range(len(index)) : phase[i] = phases[i,index[i]]
		phases 			= None
		if power: 
			pwrs 		= cw.getpower()		
			pwr 		= np.zeros(len(index))		
			for i in range(len(index)):
				pwr[i] = pwrs[index[i],i]	
			return nts.Tsd(lfp.index.values, phase), nts.Tsd(lfp.index.values, pwr)
		else:
			return nts.Tsd(lfp.index.values, phase)

#########################################################
# INTERPOLATION
#########################################################
def interpolate(z, x, y, inter, bbox = None):	
	import scipy.interpolate
	xnew = np.arange(x.min(), x.max()+inter, inter)
	ynew = np.arange(y.min(), y.max()+inter, inter)
	if bbox == None:
		f = scipy.interpolate.RectBivariateSpline(y, x, z)
	else:
		f = scipy.interpolate.RectBivariateSpline(y, x, z, bbox = bbox)
	znew = f(ynew, xnew)
	return (xnew, ynew, znew)

def filter_(z, n):
	from scipy.ndimage import gaussian_filter	
	return gaussian_filter(z, n)


#########################################################
# HELPERS
#########################################################
def writeNeuroscopeEvents(path, ep, name):
	f = open(path, 'w')
	for i in range(len(ep)):
		f.writelines(str(ep.as_units('ms').iloc[i]['start']) + " "+name+" start "+ str(1)+"\n")
		f.writelines(str(ep.as_units('ms').iloc[i]['end']) + " "+name+" end "+ str(1)+"\n")
	f.close()		
	return

def getAllInfos(data_directory, datasets):
	allm = np.unique(["/".join(s.split("/")[0:2]) for s in datasets])
	infos = {}
	for m in allm:
		path = os.path.join(data_directory, m)
		csv_file = list(filter(lambda x: '.csv' in x, os.listdir(path)))[0]
		infos[m.split('/')[1]] = pd.read_csv(os.path.join(path, csv_file), index_col = 0)
	return infos

def computeSpeed(position, ep, bin_size = 0.1):
	time_bins 	= np.arange(position.index[0], position.index[-1]+bin_size*1e6, bin_size*1e6)
	index 		= np.digitize(position.index.values, time_bins)
	tmp 		= position.groupby(index).mean()
	tmp.index 	= time_bins[np.unique(index)-1]+(bin_size*1e6)/2
	distance	= np.sqrt(np.power(np.diff(tmp['x']), 2) + np.power(np.diff(tmp['z']), 2))
	speed 		= nts.Tsd(t = tmp.index.values[0:-1]+ bin_size/2, d = distance/bin_size)
	speed 		= speed.restrict(ep)
	return speed

#######################################################
# SYNCHRONY HELPERS
#######################################################
@jit(nopython=True, parallel=True)
def getSyncAsync(spk1, spk2):
	spksync = []
	spkasync = []
	for t in spk2:
		d = np.abs(t-spk1)
		idx = d<4
		if np.sum(idx):
			spksync.append(t)
		else:
			spkasync.append(t)
	spksync = np.array(spksync)
	spkasync = np.array(spkasync)
	return spksync, spkasync


def getSpikesSyncAsync(spks, hd_neurons):
	"""
	"""		
	allsync = {}
	allasync = {}
	n = len(spks)
	for i in range(n):
		for j in range(n):
			if i != j:				
				spk1 = spks[i]
				spk2 = spks[j]
				spksync, spkasync = getSyncAsync(spk1, spk2)
				allsync[(hd_neurons[i],hd_neurons[j])] = nts.Ts(spksync, time_units = 'ms')
				allasync[(hd_neurons[i],hd_neurons[j])] = nts.Ts(spkasync, time_units = 'ms')

	return allsync, allasync

	
def sample_frates(frates_value, time_index, hd_neurons, n = 30):
	spikes_random = {}
	for i in range(len(hd_neurons)):
		l = frates_value[:,i]
		s = np.random.poisson(l, size=(int(n),int(len(frates_value))))
		tmp = {}
		for j in range(len(s)):
			tmp[j] = nts.Ts(time_index[s[j]>0])
		spikes_random[hd_neurons[i]] = tmp
	return spikes_random

def sampleSpikesFromAngularPosition(tcurve, angles, ep, bin_size = 1000):
	"""
	"""
	hd_neurons = tcurve.columns.values
	bin_size = 1000  # us
	angles = pd.Series(index = angles.index, data = np.unwrap(angles.values))
	bins = np.arange(ep.loc[0,'start'], ep.loc[0, 'end'], bin_size)
	tmp = angles.groupby(np.digitize(angles.index.values, bins)-1).mean()
	angle2 = pd.Series(index = np.arange(len(bins)), data = np.nan)
	angle2.loc[tmp.index] = tmp.values
	angle2 = angle2.interpolate()
	angle2.index = bins + np.max(np.diff(bins)/2)
	angle2 = angle2%(2*np.pi)

	idx = np.argsort(np.abs(np.vstack(angle2.values) - tcurve.index.values), axis = 1)[:,0]
	idx = tcurve.index[idx]
	frates = tcurve.loc[idx]
	frates.index = angle2.index

	time_index = frates.index.values
	frates_value = frates.values*(bin_size*1e-6)

	spikes_random = sample_frates(frates_value, time_index, hd_neurons)

	return spikes_random

def sampleSpikesFromAngularVelocity(ahvcurve, angles, ep, bin_size = 1000):
	"""
	"""
	hd_neurons = ahvcurve.columns.values
	bin_size = 1000  # us

	tmp 			= pd.Series(index = angles.index.values, data = np.unwrap(angles.values))
	tmp2 			= tmp.rolling(window=100,win_type='gaussian',center=True,min_periods=1).mean(std=10.0)
	time_bins		= np.arange(tmp.index[0], tmp.index[-1]+10000, 10000) # assuming microseconds
	index 			= np.digitize(tmp2.index.values, time_bins)
	tmp3 			= tmp2.groupby(index).mean()
	tmp3.index 		= time_bins[np.unique(index)-1]+bin_size/2
	tmp3 			= nts.Tsd(tmp3)
	tmp4			= np.diff(tmp3.values)/np.diff(tmp3.as_units('s').index.values)
	tmp2 			= nts.Tsd(tmp2)
	tmp4			= np.diff(tmp2.values)/np.diff(tmp2.as_units('s').index.values)	
	velocity 		= nts.Tsd(t=tmp2.index.values[1:], d = tmp4)
	velocity 		= velocity.restrict(ep)	
	
	bins = np.arange(ep.loc[0,'start'], ep.loc[0, 'end'], bin_size)
	tmp = velocity.groupby(np.digitize(velocity.index.values, bins)-1).mean()
	velocity2 = pd.Series(index = np.arange(len(bins)), data = np.nan)
	velocity2.loc[tmp.index] = tmp.values
	velocity2 = velocity2.interpolate()
	velocity2.index = bins + np.max(np.diff(bins)/2)
	
	idx = np.argsort(np.abs(np.vstack(velocity2.values) - ahvcurve.index.values), axis = 1)[:,0]
	idx = ahvcurve.index[idx]
	frates = ahvcurve.loc[idx]
	frates.index = velocity2.index

	time_index = frates.index.values
	frates_value = frates.values*(bin_size*1e-6)

	spikes_random = sample_frates(frates_value, time_index, hd_neurons)

	return spikes_random

def loadShankStructure(generalinfo):
	shankStructure = {}
	for k,i in zip(generalinfo['shankStructure'][0][0][0][0],range(len(generalinfo['shankStructure'][0][0][0][0]))):
		if len(generalinfo['shankStructure'][0][0][1][0][i]):
			shankStructure[k[0]] = generalinfo['shankStructure'][0][0][1][0][i][0]-1
		else :
			shankStructure[k[0]] = []
	
	return shankStructure	

def binSpikeTrain(spikes, epochs, bin_size, std=0):
	if epochs is None:
		start = np.inf
		end = 0
		for n in spikes:
			start = np.minimum(spikes[n].index.values[0], start)
			end = np.maximum(spikes[n].index.values[-1], end)
		epochs = nts.IntervalSet(start = start, end = end)

	bins = np.arange(epochs['start'].iloc[0], epochs['end'].iloc[-1] + bin_size*1000, bin_size*1000)
	rate = []
	for i,n in enumerate(spikes.keys()):
		count, _ = np.histogram(spikes[n].index.values, bins)
		rate.append(count)
	rate = np.array(rate)	
	rate = nts.TsdFrame(t = bins[0:-1]+bin_size//2, d = rate.T)
	if std:
		rate = rate.as_dataframe()
		rate = rate.rolling(window=std*20,win_type='gaussian',center=True,min_periods=1).mean(std=std)	
		rate = nts.TsdFrame(rate)
	rate = rate.restrict(epochs)
	rate.columns = list(spikes.keys())
	return rate

def shuffleByIntervalSpikes(spikes, epochs):
	shuffled = {}
	for n in spikes.keys():
		isi = []
		for i in epochs.index:
			spk = spikes[n].restrict(epochs.loc[[i]])
			tmp = np.diff(spk.index.values)
			np.random.shuffle(tmp)
			isi.append(tmp)
		shuffled[n] = nts.Ts(t = np.cumsum(np.hstack(isi)) + epochs.loc[0,'start'])
	return shuffled

# def plotTuningCurves(tcurves, tokeep = []):	
# 	figure()
# 	for i in tcurves.columns:
# 		subplot(int(np.ceil(np.sqrt(tcurves.shape[1]))),int(np.ceil(np.sqrt(tcurves.shape[1]))),i+1, projection='polar')
# 		plot(tcurves[i])
# 		if len(tokeep):
# 			if i in tokeep:
# 				plot(tcurves[i], linewidth = 4)
# 	return


######################################################################################
# OPTO STUFFS
######################################################################################
def computeRasterOpto(spikes, opto_ep, bin_size = 100):
	"""
	Bin size in ms
	edge in ms
	"""
	rasters = {}
	frates = {}

	# assuming all opto stim are the same for a session
	stim_duration = opto_ep.loc[0,'end'] - opto_ep.loc[0,'start']
	
	bins = np.arange(0, stim_duration + 2*stim_duration + bin_size*1000, bin_size*1000)

	for n in spikes.keys():
		rasters[n] = []
		r = []
		for e in opto_ep.index:
			ep = nts.IntervalSet(start = opto_ep.loc[e,'start'] - stim_duration,
								end = opto_ep.loc[e,'end'] + stim_duration)
			spk = spikes[n].restrict(ep)
			tmp = pd.Series(index = spk.index.values - ep.loc[0,'start'], data = e)
			rasters[n].append(tmp)
			count, _ = np.histogram(tmp.index.values, bins)
			r.append(count)
		r = np.array(r)
		frates[n] = pd.Series(index = bins[0:-1]/1000, data = r.mean(0))
		rasters[n] = pd.concat(rasters[n])		

	frates = pd.concat(frates, 1)
	frates = nts.TsdFrame(t = frates.index.values, d = frates.values, time_units = 'ms')
	return frates, rasters, bins, stim_duration



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
        'pandas>=1.0.3',
        'numba>=0.46.0',
        'numpy',
        'scipy'

]

test_requirements = [ ]

setup(
    author="Guillaume Viejo",
    author_email='guillaume.viejo@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="PYthon Neural Analysis Package Pour Laboratoires d’Excellence",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description='pynapple is a Python library for analysing neurophysiological data. It allows to handle time series and epochs but also to use generic functions for neuroscience such as tuning curves and cross-correlogram of spikes. It is heavily based on neuroseries.' 
    + '\n\n' + history,
    include_package_data=True,
    keywords='neuroscience',
    name='pynapple',
    packages=find_packages(include=['pynapple', 'pynapple.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/PeyracheLab/pynapple',
    version='0.1.1.post1',
    zip_safe=False,
    long_description_content_type='text/x-rst'
)
